#include <stddef.h>
#include "ljp_pnoload.h"
#include "host68k.h"
#include "pnobridge.h"
#include <MemGlue.h>

#define EndianSwap32(n)	(((((UInt32) n) << 24) & 0xFF000000) |	\
                         ((((UInt32) n) <<  8) & 0x00FF0000) |	\
                         ((((UInt32) n) >>  8) & 0x0000FF00) |	\
                         ((((UInt32) n) >> 24) & 0x000000FF))

#define ALIGN4(addr) (((UInt32)(addr) + 3) & 0xFFFFFFFC)



typedef struct PnoEntryHeader {
	UInt32 r10Value;
	UInt32 pnoMainAddress;
	UInt32 userDataP;
} PnoEntryHeader;

// this is the startup code defined in CW's ARMlet_Startup.c with a table of
// offsets into the PNO following the startup code

typedef struct PnoHeaderType {
	UInt32 startupCode[8];	// changes based on the instruction count in startup code 
	UInt32 pnoMain;         // offset to ARMletMain routine
	UInt32 signature;       // special PNO signature value
	UInt32 dataStart;       // offset to start of initialized data
	UInt32 roDataStart;     // offset to start of read-only initialized data
	UInt32 bssStart;        // offset to start of uninitialized data
	UInt32 bssEnd;          // offset to end of uninitialized data
	UInt32 codeRelocsStart; // offset to start of data-to-code relocation list
	UInt32 codeRelocsEnd;   // offset to end of data-to-code relocation list
	UInt32 dataRelocsStart; // offset to start of data-to-data relocation list
	UInt32 dataRelocsEnd;   // offset to end of data-to-data relocation list
	UInt32 altEntryCode[8];	// changes based on the instruction count in alternate entry code 
} PnoHeaderType;


static void ljp_PnoProcessCWRelocationList(
	UInt32 * listStartP, UInt32 * listEndP, MemPtr dataBaseP, UInt32 addend)
{
	// relocation list is a simple list of UInt32 offsets into the PNO that will
	// be adjusted by adding in the relocated start of code address or R10 value.

	UInt32 * currentRelocP = listStartP;
	while (currentRelocP != listEndP) {
		UInt32 offset = EndianSwap32(*currentRelocP);
		UInt32 * relocLocation = (UInt32*)((char*)dataBaseP + offset);
		UInt32 value = EndianSwap32(*relocLocation) + addend;
		*relocLocation = EndianSwap32(value);
		++currentRelocP;
	}
}


static Err ljp_PnoLoadCodeWarrior(PnoDescriptor * descriptorP, MemPtr pnoP, UInt32 pnoSize)
{
	PnoHeaderType * pnoHeaderP = (PnoHeaderType*)pnoP;
	UInt32 pnoMain;

	MemPtr dataP;
	UInt32 newR10Value;
	
	UInt32 bssStart;
	UInt32 dataStart;
	UInt32 bssEnd;

	UInt32 codeRelocsStart;
	UInt32 codeRelocsEnd;
	UInt32 dataRelocsStart;
	UInt32 dataRelocsEnd;
	
	if (pnoP == NULL) return memErrInvalidParam;
	
	pnoHeaderP      = (PnoHeaderType*)pnoP;
	pnoMain         = EndianSwap32(pnoHeaderP->pnoMain);
	bssStart        = EndianSwap32(pnoHeaderP->bssStart);
	dataStart       = EndianSwap32(pnoHeaderP->dataStart);
	bssEnd          = EndianSwap32(pnoHeaderP->bssEnd);
	codeRelocsStart = EndianSwap32(pnoHeaderP->codeRelocsStart);
	codeRelocsEnd   = EndianSwap32(pnoHeaderP->codeRelocsEnd);
	dataRelocsStart = EndianSwap32(pnoHeaderP->dataRelocsStart);
	dataRelocsEnd   = EndianSwap32(pnoHeaderP->dataRelocsEnd);

	if (bssStart < dataStart || bssEnd < bssStart) {
		DEBUGS("Invalid PNO.");
		return memErrInvalidParam;
	}

//	pnoSize = MemPtrSize(pnoP);

	// .bss section may not in the image but we don't care
	if (pnoSize < bssStart) {
		DEBUGS("Wrong size of PNO.");
		return memErrInvalidParam;
	}


	{
		MemPtr new_pnoP = MemGluePtrNew(pnoSize);
		if (new_pnoP == NULL) {
			DEBUGS("Out of memory loading PNO.");
			return memErrInvalidParam;
		}												
		MemMove(new_pnoP,pnoP,pnoSize);
		MemSet((void*)((UInt32)new_pnoP+bssStart),bssEnd-bssStart,0);

		pnoP=new_pnoP;			
		dataP=(MemPtr)((UInt32)pnoP+dataStart);
	}
	
	pnoHeaderP      = (PnoHeaderType*)pnoP;
	pnoMain         = EndianSwap32(pnoHeaderP->pnoMain);
	bssStart        = EndianSwap32(pnoHeaderP->bssStart);
	dataStart       = EndianSwap32(pnoHeaderP->dataStart);
	bssEnd          = EndianSwap32(pnoHeaderP->bssEnd);
	codeRelocsStart = EndianSwap32(pnoHeaderP->codeRelocsStart);
	codeRelocsEnd   = EndianSwap32(pnoHeaderP->codeRelocsEnd);
	dataRelocsStart = EndianSwap32(pnoHeaderP->dataRelocsStart);
	dataRelocsEnd   = EndianSwap32(pnoHeaderP->dataRelocsEnd);
	
	

	newR10Value = (UInt32)pnoP;


	// Perform data-to-code relocations
	ljp_PnoProcessCWRelocationList(
		(UInt32*)((UInt32)pnoP + codeRelocsStart), 
		(UInt32*)((UInt32)pnoP + codeRelocsEnd), 
		(MemPtr)newR10Value,
		(UInt32)pnoP/*newR10Value*/);

	// Perform data-to-data relocations
	ljp_PnoProcessCWRelocationList(
		(UInt32*)((UInt32)pnoP + dataRelocsStart), 
		(UInt32*)((UInt32)pnoP + dataRelocsEnd), 
		(MemPtr)newR10Value, 
		(UInt32)newR10Value);
	
	// setup PNO descriptor so we can call the PNO and unload it when done
	descriptorP->entry = (NativeFuncType *)&(pnoHeaderP->altEntryCode);
	descriptorP->dataP = dataP;
	descriptorP->r10Value = EndianSwap32(newR10Value);
	descriptorP->pnoMainAddress = EndianSwap32((UInt32)pnoP + pnoMain);
	descriptorP->creator = 0;
	descriptorP->ftrNum = 0;

	return errNone;
}

Err ljp_PnoUnload(PnoDescriptor * descriptorP)
{
	Err err = errNone;
	
	if (descriptorP == NULL) {
		DEBUGS("NULL passed for PNO descriptor address.");
		return memErrInvalidParam;
	}

	if (descriptorP->creator != 0) {
		err = err || FtrPtrFree(descriptorP->creator, descriptorP->ftrNum);
		descriptorP->creator = 0;
		descriptorP->ftrNum = 0;
	}
	
	if (descriptorP->r10Value != NULL)
		MemPtrFree((MemPtr )EndianSwap32(descriptorP->r10Value));

	ErrNonFatalDisplayIf(err != errNone, "Error unloading PNO.");
	return err;
}


Err ljp_PnoLoad(PnoDescriptor * descriptorP, MemPtr pnoP, UInt32 pnoSize)
{
	UInt32 signature;

	if (pnoP == NULL) {
		DEBUGS("PNO pointer is NULL")
		return memErrInvalidParam;
	}	
	
	signature = EndianSwap32(((PnoHeaderType*)pnoP)->signature);

	if (signature == 'cdwr') {
		return ljp_PnoLoadCodeWarrior(descriptorP, pnoP,pnoSize);
	}
	else {
		DEBUGS("Unrecognized PNO format.");
		return memErrInvalidParam;
	}
}
UInt32 ljp_PnoCall(PnoDescriptor * descriptorP, void *userDataP)
{
	char headerBuffer[sizeof(PnoEntryHeader) + 2];
	PnoEntryHeader * header = (PnoEntryHeader *)ALIGN4(headerBuffer);

	header->r10Value       = descriptorP->r10Value;
	header->pnoMainAddress = descriptorP->pnoMainAddress;
	header->userDataP      = EndianSwap32((UInt32)userDataP);

	return PceNativeCall(descriptorP->entry, header);
}
